/*
** avfilter.c - written in milano by vesely on 19 oct 2001
** sophos anti virus for courier-mta global filters
*/
/*
** Copyright (c) 1999, 2000, 2001, 2002, 2003, 2004 Alessandro Vesely
** All rights reserved. see COPYING
*/
#if defined(HAVE_CONFIG_H)
#include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#if HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <assert.h>
#include <errno.h>
#include <ctype.h>
#include <limits.h>

#include <sys/types.h>
#if TIME_WITH_SYS_TIME
#include <sys/time.h>
#include <time.h>
#else
#if HAVE_SYS_TIME_H
#include <sys/time.h>
#else
#include <time.h>
#endif
#endif

#include <sys/socket.h>
#include <sys/un.h>
#include <sys/wait.h>
#if HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif
#include <signal.h>
#if HAVE_FCNTL_H
#include <fcntl.h>
#endif
#include <sys/mman.h>

/* COURIER */
#include "unconfig.h"
#include <courier/filtersocketdir.h>
#include <courier/courier.h> /* for config_localfilename */
#include <libfilter/libfilter.h>

#if defined(HAVE_CONFIG_H)
#include "unconfig.h"
#include <config.h>
#endif


/* SAV */
#include <sophos/compute.h>
#include <sophos/savitype.h>
/*
 * -ale: INITGUID is the declare/define interface switch
 */
#define INITGUID
#include <sophos/csavi3c.h>
#undef INITGUID

/* MIMEC */
#include "mime.h"
#include "parse.h"

/* filter_msg lib */
#include "filter_msg.h"
#include "filterlib.h"
#include "avfiledefs.h"

#if !defined(PATH_MAX)
#if defined(MAX_PATH)
#define PATH_MAX MAX_PATH
#else
#define PATH_MAX 255
#endif
#endif

/* forward declaration of struct and message function */
struct avfilter_parm_struct;
typedef struct avfilter_parm_struct av_parm;
static int virus_detected(av_parm *parm, filter_msg *vm,
	char const *virus_name, char const*location);

typedef struct virus_load_stats
{
	time_t date;
	unsigned count;
} virus_load_stats;

static const char *month_abbr(unsigned ndx)
{
	static const char *abbr[] =
	{"Jan", "Feb", "Mar", "Apr", "May", "Jun",
	 "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};
	if (ndx >= sizeof abbr / sizeof abbr[0]) return "---";
	return abbr[ndx];
}

int SAVI_load_n_report(CISavi3* pSAVI, FILE *report,
	int verbose, virus_load_stats *virus_load)
{
	HRESULT hr;
	CIEnumIDEDetails *IDElist = NULL;
	SYSTEMTIME date;
	OLECHAR vdataVerString[101];
	U32 virNum, vengVer,
		vdataVerStringLength =
			sizeof vdataVerString / sizeof vdataVerString[0] - 1;
	time_t start, end;
	int rtc = 0;
	
	/*
	* Load the virus data (if called at startup 
	* this should be done before auto-load)
	*/
	hr = pSAVI->pVtbl->LoadVirusData(pSAVI);
	if (SOPHOS_FAILED(hr))
	{
		fprintf(stderr,
			"ERR: avfilter: LoadVirusData error: %#lx\n",
			(unsigned long) hr);
		if (stderr != report)
			fprintf(report,
				"avfilter: LoadVirusData error: %#lx\n"
				"          (perhaps virus data too old?)",
				(unsigned long) hr);
		rtc = 1;
	}
	
	/*
	* Ask for the engine information - virus engine and data version number,
	* and the number of viruses.
	*/
	time(&start);
	hr = pSAVI->pVtbl->GetVirusEngineVersion(pSAVI,
		&vengVer,
		(LPOLESTR) vdataVerString,
		vdataVerStringLength,
		&date,
		&virNum,
		NULL,
		(REFIID)&SOPHOS_IID_ENUM_IDEDETAILS,
		(void **)&IDElist);
	time(&end);
	
	if (SOPHOS_FAILED(hr))
	{
		fprintf(report,
			"ALERT: avfilter: Error %#lx getting version number from engine\n",
			(unsigned long)hr);
		rtc = 1;
	}
	else
	{
		if (verbose > 1)
		{
			if (end > start)
				fprintf(report,
					"avfilter: virus data loaded in: %d sec(s)\n",
					(int)(end - start));
			if (virus_load->date)
			{
				struct tm tm = *localtime(&virus_load->date);
				unsigned elapsed = start - virus_load->date;
				unsigned dd, hh, mm;
				char buf[32], *p = &buf[0];
				
				dd = elapsed / (24UL * 3600UL);
				if (dd)
				{
					elapsed -= dd * 24UL * 3600UL;
					p += sprintf(p, "%u day(s) ", dd);
				}
				hh = elapsed / 3600UL;
				if (hh || dd)
				{
					elapsed -= hh * 3600UL;
					p += sprintf(p, "%uh ", hh);
				}
				mm = elapsed / 60UL;
				elapsed -= mm * 60UL;
				sprintf(p, "%um ", mm);
				
				fprintf(report,
					"avfilter: previous load was:    %d %s %d (%s%us ago)\n",
					tm.tm_mday, month_abbr(tm.tm_mon), tm.tm_year + 1900,
					buf, elapsed);
			}
			fprintf(report, "avfilter: Virus engine version: %u.%u\n", 
				(unsigned int) ((vengVer & 0xFFFF0000) >> 16),
				(unsigned int) (vengVer & 0x0000FFFF) );
			fprintf(report, "avfilter: Virus data version:   %s\n",
				vdataVerString);
			fprintf(report, "avfilter: Virus data date:      %d %s %d\n",
				date.wDay, month_abbr(date.wMonth - 1), date.wYear);
		}
		if (verbose > 0)
			fprintf(report, "avfilter: No. of viruses:       %ld\n",
				(unsigned long)virNum);
		if (verbose > 1 && virus_load->count)
			fprintf(report, "avfilter: previous count was:   %u (%d increment)\n",
				virus_load->count, (int)(virNum - virus_load->count));
		virus_load->count = virNum;
		virus_load->date = start;
	}
	
	/*
	* Release enumerator interface (hmmm... we didn't use it)
	*/
	if (IDElist)
		IDElist->pVtbl->Release(IDElist);
	
	return rtc;
}

static int ListInfections(CIEnumSweepResults *pEnumResults,
	av_parm *parm, filter_msg *vm, char const *location)
{
  const U32       BufLen     = 80;
  CISweepResults *pResults   = NULL;
  HRESULT         hr;
  OLECHAR         Buf[81];
  SOPHOS_ULONG    pcFetched;
  
  /*
   * Sanity.
   */
  if ((!location) || (!vm) || (!pEnumResults))
  {
    fprintf(stderr,
			"ERR: avfilter: ListInfections() called with incorrect parameters.\n");
    return 1;
  }

  /*
   * Ensure that the enumerator is positioned
   * at the start.  If this fails, return now.
   */
  hr = pEnumResults->pVtbl->Reset(pEnumResults);
  if (SOPHOS_FAILED(hr))
  {
    fprintf(stderr, "ERR: avfilter: Failed to reset enumerator [0x%08lx].\n",
    	(unsigned long)hr);
    return 1;
  }

  /*
   * Walk through the enumeration interface,
   * retrieving one results object at a time.
   */
	while (pEnumResults->pVtbl->Next(pEnumResults,
		1, (void **)&pResults, &pcFetched) == SOPHOS_S_OK)
	{
		hr = pResults->pVtbl->GetVirusName(pResults,
			BufLen, (LPOLESTR) Buf, NULL);
		if (SOPHOS_FAILED(hr))
		{
			fprintf(stderr,
				"ERR: avfilter: Failed to get virus name [0x%08lx].\n",
				(unsigned long)hr);
			strcpy(Buf, "Unknown");
		}
		
		if (virus_detected(parm, vm, Buf, location))
			goto bail;
		
		pResults->pVtbl->Release(pResults);
		pResults = NULL;
  }
  
  return 0;

bail:
  if (pResults)
    pResults->pVtbl->Release(pResults);
  return 1;
}


/*
 * Function: InitialiseSAVI
 *
 * Parameters: (1) pointer to the SAVI client name.
 *
 * Returns: Initialised SAVI interface ready for use
 *          on success.
 *          NULL on failure.
 *
 * Purpose: Performs the bootstrap process -- creates
 *          the SAVI interface from scratch, and
 *          initialises it ready for use.
 */
static CISavi3 *InitialiseSAVI(LPCOLESTR ClientName)
{
  CISavi3              *pSAVI = NULL;
  CISweepClassFactory2 *pFactory =NULL;
  HRESULT               hr;
    
  /*
   * Request a class factory interface.
   */
  hr = DllGetClassObject((REFIID)&SOPHOS_CLASSID_SAVI,
                         (REFIID)&SOPHOS_IID_CLASSFACTORY2,
                         (void **)&pFactory);
  if (SOPHOS_FAILED(hr))                                                                          
  {
    fprintf(stderr, "ALERT: Failed to get class factory interface [0x%08lx].\n",
    	(unsigned long)hr);
    return NULL;
  }

  /*
   * Ask the class factory for a CSAVI3 interface.
   */
  hr = pFactory->pVtbl->CreateInstance(pFactory,
                                       NULL,
                                       &SOPHOS_IID_SAVI3,
                                       (void **)&pSAVI);

  /*
   * Drop the factory immediately, we don't need it
   * again in this example.
   */
  pFactory->pVtbl->Release(pFactory);

  /*
   * Did we get the interface we requested?
   */
  if (SOPHOS_FAILED(hr))
  {
    fprintf(stderr, "ALERT: Failed to get a CSAVI3 interface [0x%08lx].\n",
    	(unsigned long)hr);
    return NULL;
  }

  /*
   * Ask SAVI to initialise itself.
   */
	hr = pSAVI->pVtbl->InitialiseWithMoniker(pSAVI, ClientName);

  /*
   * If the initialisation failed, then release the SAVI
   * interface and set the pointer to NULL.
   */
  if (SOPHOS_FAILED(hr))
  {
    fprintf(stderr, "ALERT: Failed to initialise SAVI [0x%08lx].",
    	(unsigned long)hr);
    pSAVI->pVtbl->Release(pSAVI);
    pSAVI = NULL;
  }

  return pSAVI;
}

static int get_engine_option(CISavi3 *pSAVI, char const *name, int verbose)
/*
** get a true/false (0/1 type 3) option
*/
{
	char buf[80];
	U32 size = sizeof buf / 2;
	HRESULT hr = pSAVI->pVtbl->
		GetConfigValue(pSAVI, name, 3, size, buf, &size);
	if (SOPHOS_FAILED(hr) || size > 2)
	{
		if (verbose > 2)
			fprintf(stderr,
				"avfilter: cannot get %s[3]: size=%u [0x%08lx]\n",
				name, (unsigned int)size, (unsigned long)hr);
		return -1;
	}
	return buf[0] - '0';
}

static int
set_engine_option(CISavi3 *pSAVI, char const *name, int verbose, int value)
/*
** set a true/false (0/1 type 3) option
*/
{
	int rtc = 0;
	HRESULT hr;
	char buf[80];
	
	sprintf(buf, "%d", value);
	hr = pSAVI->pVtbl->
		SetConfigValue(pSAVI, name, 3, buf);
	if (SOPHOS_FAILED(hr) || strlen(buf) != 1)
	{
		if (verbose > 2)
			fprintf(stderr,
				"avfilter: cannot set %s[3]=%s: [0x%08lx]\n",
				name, buf, (unsigned long)hr);
		rtc = 1;
	}
	return rtc;
}

/*
 * Function: DescribeConfig
 *
 * Parameters: (1) initialised SAVI interface.
 *             (2) configuration interface to describe.
 *
 * Returns: 0 on failure, !0 on success.
 *
 * Purpose: Displays details of the given configuration interface.
 */
static int DescribeConfig(CISavi3 *pSAVI, CIEngineConfig *pConfig)
{
  const U32 arraySize = 100;
  U32       type      = 0;
  OLECHAR   name[100];
  OLECHAR   value[100];
  HRESULT   hr;

  /*
   * Sanity check.
   */
  if ((!pSAVI) || (!pConfig))
  {
    fprintf(stderr,
       "ERR: DescribeConfig() called with incorrect parameters.\n");
    return 1;
  }

  /*
   * Initialise locals.
   */
  name[0]  = (OLECHAR)0;
  value[0] = (OLECHAR)0;
  type     = 0;

  /*
   * First, get the name of this configuration object.
   */
  hr = pConfig->pVtbl->GetName(pConfig, arraySize, name, NULL);
  if (SOPHOS_FAILED(hr))
  {
    fprintf(stderr, "ERR: Couldn't get option name [0x%08lx].\n",
               (unsigned long)hr);
    return 1;
  }

  /*
   * Now, get the type of this configuration object.
   */
  hr = pConfig->pVtbl->GetType(pConfig, &type);
  if (SOPHOS_FAILED(hr))
  {
    fprintf(stderr, "ERR: Couldn't get size [0x%08lx].\n", (unsigned long)hr);
    return 1;
  }

  /*
   * Use the SAVI interface to get the current value.
   */
  hr = pSAVI->pVtbl->GetConfigValue(pSAVI, name, type, arraySize, value, NULL);
  if (SOPHOS_FAILED(hr))
  {
    fprintf(stderr, "ERR: Couldn't get current value [0x%08lx].\n", 
               (unsigned long)hr);
    return 1;
  }

  /*
   * Finally, display the information we've gathered.
   */
  fprintf(stdout, "set %-26s [%ld] = %s\n", name, (long)type, value);

  /*
   * Return successfully.
   */
  return 0;
}

/* ----- */

/*
 * Function: ListConfigs
 *
 * Parameters: (1) pointer to the SAVI interface.
 *
 * Returns: 0 on failure, !0 on success.
 *
 * Purpose: Given an initialised SAVI interface, display 
 *          information about the available configuration 
 *          settings (and their current values).
 */
static int ListConfigs(CISavi3 *pSAVI)
{
  CIEnumEngineConfig *pConfigEnum = NULL;
  CIEngineConfig     *pConfig     = NULL;
  U32                 pcFetched;
  HRESULT             hr;
  int rtc = 0;

  /*
   * Sanity check.
   */
  if (!pSAVI)
  {
    fprintf(stderr, "ERR: ListConfigs() called with incorrect parameters.\n");
    return 1;
  }

  /*
   * Ask the SAVI interface for a configuration enumeration interface.  
   * If it fails, bail out.
   */
  hr = pSAVI->pVtbl->GetConfigEnumerator(pSAVI,
                                         (REFIID)&SOPHOS_IID_ENUM_ENGINECONFIG,
                                         (void **)&pConfigEnum);
  if (SOPHOS_FAILED(hr))
  {
    fprintf(stderr, "ERR: Couldn't get ConfigEnumerator [0x%08lx].\n",
               (unsigned long)hr);
    return 1;
  }

  /*
   * Now we have a configuration interface.
   * Reset the enumerator to the start of the list.
   */
  hr = pConfigEnum->pVtbl->Reset(pConfigEnum);
  if (SOPHOS_FAILED(hr))
  {
    fprintf(stderr,
       "ERR: Failed to reset configuration enumerator [0x%08lx].\n",
         (unsigned long)hr);
    return 1;
  }

  /*
   * Loop through the enumerator, retrieving
   * one configuration interface at a time.
   */
  while (pConfigEnum->pVtbl->Next(pConfigEnum, 1, (void **)&pConfig, &pcFetched)
    == SOPHOS_S_OK)
  {
    /*
     * Output some information about this configuration object.
     */
    rtc |= DescribeConfig(pSAVI, pConfig);

    /*
     * Release this configuration interface
     * ready to get the next one.
     */
    pConfig->pVtbl->Release(pConfig);
  }

  /*
   * Now we can release the enumerator as well.
   */
  pConfigEnum->pVtbl->Release(pConfigEnum);

  return rtc;
}

/******************************************************************************/
typedef struct savi_scan_error_struct
{
	HRESULT hr;
	char const *msg; /* printed by filter_msg_add(..., 0) */
	filter_msg_action action;
} savi_se;

struct avfilter_parm_struct
{
	int err; /* error parsing config file */
	
	/* strings in PARSE */
	char *save_missed_file;
	char *save_virus_file;
	char *tmp_dir;
	savi_se *savi_se_table;
	unsigned savi_se_table_size;
	
	PARSE *parse;
	FILE *input_fp;
	
	/* config_file is retained upon reloading */
	char *config_file;
	FILE *report;
	
	/* other values retained on reloading */
	CISavi3 *pSAVI;
	virus_load_stats virus_load;
	
	unsigned pid_created:1;
	/* other bit fields not retained */
	unsigned scan_malformed:1;
	unsigned eng_has_Mime:1;
	unsigned eng_has_Base64:1;
	unsigned verbose:4;
	
	unsigned x55aa5a5a;
	filter_msg_action action_default;
};

static int parm_init(av_parm *parm, av_parm *old, char *fname)
/*
** initialize parameters, loading AV software. If the SAVI is already
** present in old, just copy its value. SAVI is set to config default
** only if no error occurred opening config file.
*/
{
	memset(parm, 0, sizeof *parm);
	if (old == NULL || old->pSAVI == NULL)
	{
		if ((parm->pSAVI = InitialiseSAVI("avfilter")) == NULL)
			return 1;
	}
	else
	{
		parm->pSAVI = old->pSAVI;
	}
	
	if (old)
	{
		parm->pid_created = old->pid_created;
		parm->virus_load = old->virus_load;
	}
	
	if (fname == NULL)
		fname = parm->config_file = config_localfilename("filters/avfilter.conf");
	else
		fname = parm->config_file = strdup(fname);
	if (fname == NULL)
		parm->err = 1;
	else if ((parm->parse = parse_init(fname)) == NULL)
		parm->err = (errno == ENOENT ? 0 : 1);

	if (parm->err == 0)
	{
		CISavi3 *const pSAVI = parm->pSAVI;
		pSAVI->pVtbl->SetConfigDefaults(pSAVI);
	}
	parm->action_default = msg_drop;
	parm->x55aa5a5a = 0x55aa5a5a;
	return parm->err;
}

typedef enum parm_clear_option
{
	clear_all, clear_all_but_SAVI
} parm_clear_option;

static void parm_clear(av_parm *parm, parm_clear_option opt)
{
	CISavi3 *const pSAVI = parm->pSAVI;
	if (pSAVI && opt == clear_all)
	{
		pSAVI->pVtbl->Terminate(pSAVI);
		pSAVI->pVtbl->Release(pSAVI);
	}
	if (parm->config_file)
		free(parm->config_file);
	if (parm->savi_se_table)
		free(parm->savi_se_table);
	
#if 0
	/* directories kept in parse buffer */
	if (parm->save_virus_file)
		free(parm->save_virus_file);
	if (parm->save_missed_file)
		free(parm->save_missed_file);
	if (parm->tmp_dir)
		free(parm->tmp_dir);
#endif

	if (parm->parse)
		parse_clear(parm->parse);

	// parm->report is not closed
	memset(parm, 0, sizeof *parm);
}

static void save_file(av_parm *parm, int const virus, char const *const name)
{
	static const char templ[] = "/avf-";
	char const *dir = virus ? parm->save_virus_file : parm->save_missed_file;
	unsigned const dir_len = strlen(dir);
	unsigned const name_len = strlen(name);
	unsigned const namesize = dir_len + sizeof templ + name_len + 10;
	char *const fname = (char*)malloc(namesize);
	char const *error;
	int fno /* compiler happy */ = 0;
	
	if (fname == NULL)
		error = "malloc fail";
	else
	{
		char *p = fname + dir_len + sizeof templ - 1,
		// max space available for name; 6 for XXXXXX
			*const end = fname + PATH_MAX - 6;
		
		memcpy(fname, dir, dir_len);
		memcpy(fname + dir_len, templ, sizeof templ);
		
		if (p > end)
			error = "path too long";
		else
		{
			unsigned char const *s = name;
			while (p < end)
			{
				int ch = *s++;
				if (ch == 0)
					break;
				
				if (!isalnum(ch))
					ch = '_';
				*p++ = ch;
			}
			if (p < end)
				*p++ = '-';
			//         123456
			strcpy(p, "XXXXXX");
			assert(strlen(fname) == dir_len + sizeof templ - 1
				+ (name_len < (PATH_MAX - (dir_len + sizeof templ - 1) - 6) ?
					name_len : (PATH_MAX - (dir_len + sizeof templ - 1) - 6))
				+ (name_len < (PATH_MAX - (dir_len + sizeof templ - 1) - 6) ? 1 : 0)
				+ 6);
			assert(strlen(fname) < namesize);

			fno = mkstemp(fname);
			if (fno == -1)
				error = "mkstemp fails";
			else
				error = NULL;
		}
	}
	
	if (error)
	{
		fprintf(stderr, "ERR: avfilter: Unable to save %.*s: %s\n",
			PATH_MAX, fname ? fname: "(null)", error);
	}
	else
	{
		FILE *fps = fdopen(fno, "w+");
		FILE *const fp = parm->input_fp;
		char buf[4096];
		size_t r = 1;
		if (fps)
		{
			fseek(fp, 0, SEEK_SET);
			while ((r = fread(buf, 1, sizeof buf, fp)) > 0)
				if (fwrite(buf, r, 1, fps) != 1)
					break;
			r = (ferror(fp) || ferror(fps)) ? 1 : 0;
			fclose(fps);
		}
		else
			close(fno);
		if (r)
		{
			fprintf(stderr, "ERR: avfilter: Unable to save %s: %s\n",
				fname, strerror(errno));
			unlink(fname);
		}
		else // all ok, don't save this message anymore (for this run)
		{
			parm->save_virus_file = parm->save_missed_file = NULL;
			if (parm->verbose > 1)
				fprintf(stderr, "avfilter: Saved %s file in %s\n",
					virus ? "virus" : "missed", fname);
		}
	}
	free(fname);
}

static int colon_first(char const *p, size_t size)
/* return 1 if a colon is the first non-white space in p */
{
	while (size > 0)
	{
		int const ch = *(unsigned char const*)p;
		if (ch == ':')
			return 1;
		if (!isspace(ch))
			return 0;
		++p;
		--size;
	}
	return 0;
}

static int is_discard_header(char const *head, size_t size)
/* return 1 if the header is to, cc or bcc */
{
	static char const * const discard_header[] = { "to", "cc", "bcc" };
	unsigned i;

	for (i = 0; i < sizeof discard_header / sizeof discard_header[0]; ++i)
	{
		char const * const p = discard_header[i];
		unsigned len = strlen(p);
		if (len < size &&
			strincmp(head, p, len) == 0 &&
			colon_first(p + len, size - len))
				return len;
	}
	return 0;
}

static int read_message_header(FILE *fp, char *buf, size_t size)
/*
** read one header into a large enough buffer, discarding to, cc, bcc.
** return -2 = i/o error, -1 = length error or discarded header
** 0 = end of headers, len = length of header read
*/
{
	int flag = 1;
	unsigned len = 0;
	
	assert(size >= 1024);
	
	while (len < size && fgets(&buf[len], size - len, fp) != NULL)
	{
		int ch;
		
		len = strlen(buf);
		if (len <= 0 || buf[len - 1] != '\n') /* bogus */
		{
			flag = -1;
			len = 0;
		}
		
		else if (len == 1) /* end of headers */
			return 0;
			
		else /* fine, check for wrapped header */
		{
			ch = fgetc(fp);
			if (ch == '\n' || !isspace(ch))
			{
				ungetc(ch, fp);
				buf[--len] = 0;
				return len;
			}
			buf[len++] = ch;
		
			if (flag)
			/* on first header line check if discarding this
			** for discarded headers, always reset length
			*/
			{
				if (flag < 0 || is_discard_header(buf, len))
				{
					flag = -1;
					len = 0;
				}
				else
					flag = 0;
			}
		}
	}
	assert(len < size);
	return -2;
}

static char const* read_message(av_parm *parm, MIME *mime, FILE *fp)
{
	char buf[4096];
	int rtc, mime_headers = 0;
	
	while ((rtc = read_message_header(fp, buf, sizeof buf)) > 0)
	{
		int x = mime_set_header(mime, mime_header_unknown, buf);
		if (x < 0)
		{
			rtc = -2;
			break;
		}
		if (x > 0)
			++mime_headers;
		/*
		** add here a check for a already-filtered content
		*/
	}
	
	if (rtc == 0 && mime_headers == 0 /* is text/plain by default */
	|| rtc == -2 && ferror(fp) == 0) /* only headers given */
		return "200 Ok.\n"; /* nothing to scan */
	
	if (rtc == 0)
	{
		errno = 0;
		rtc = mime_read(mime, fp);
		if (rtc == 0)
			return NULL; /* good return */
	}

	if (errno)
		fprintf(stderr, "ERR: avfilter[%lu]: MIME: %s: %s\n",
			(unsigned long) getpid(),
			mime_get_last_error_string(mime),
			strerror(errno));
	else if (parm->verbose > 1)
		fprintf(stderr, "avfilter[%lu]: MIME: %s\n",
			(unsigned long) getpid(), mime_get_last_error_string(mime));

	if (parm->save_missed_file && !ferror(fp))
		save_file(parm, 0, mime_get_last_error_string(mime));

	// this shouldn't be done here
	if (parm->scan_malformed)
		return "";
	
	if (rtc == -2)
		rtc = ferror(fp) ? MIME_INFILE_ERROR : MIME_HEADER_ERROR;
	else
		rtc = mime_get_last_error(mime);

	switch (rtc)
	{
		case MIME_INFILE_ERROR:
		case MIME_SYSFILE_ERROR:
		case MIME_SYSTEM_ERROR:
			return "451 system failure in avfilter\n";
		
		case MIME_HEADER_ERROR:
		case MIME_DATA_ERROR:
			return "554 malformed MIME structure\n";
		
		default:
			return "500 unknown server error\n";
	}	
}

static int savi_scan_error(av_parm const *parm, HRESULT hr,
	filter_msg *vm, char const *location)
{
	int i, rtc = 0;
	int const table_size = parm->savi_se_table_size;
	savi_se const* const table = parm->savi_se_table;
	
	for (i = 0; i < table_size; ++i)
	{
		savi_se const * const se = &table[i];
		if (hr == se->hr)
		{
			char const *msg = se->msg;
			if (msg == NULL)
				msg = "cannot scan content";
				
			rtc = filter_msg_add(vm, msg, location, se->action);
			break;
		}
	}
	
	if (i >= table_size)
	{
		rtc = filter_msg_add(vm, "cannot scan content", location, msg_bounce);
		fprintf(stderr, "avfilter: Cannot scan %s: [0x%08lx].\n",
			location, (unsigned long)hr);
	}
	return rtc;
}

static int virus_detected(av_parm *parm, filter_msg *vm,
	char const *virus_name, char const*location)
{
	/*
	** add here action adjustment, e.g.
	if (virus)
	{
		item.key = name;
		x = hsearch(item, FIND);
		if (x)
			action = (filter_msg_action)(x->data);
	}
	*/
	
	if (parm->save_virus_file)
		save_file(parm, 1, virus_name);
	else if (parm->verbose > 1)
		fprintf(stderr, "avfilter[%lu]: found %s in %s\n",
			(unsigned long) getpid(), virus_name, location);

	return filter_msg_add(vm, virus_name, location, parm->action_default);
}

static int scan_memory(av_parm *parm, filter_msg *vm,
	char const *location, content_ptr const *part)
{
	CIEnumSweepResults *pEnumResults = NULL;
	CISavi3 *const pSAVI = parm->pSAVI;
	HRESULT hr = pSAVI->pVtbl->SweepBuffer(pSAVI, location,
		part->length, part->start,
		(REFIID)&SOPHOS_IID_ENUM_SWEEPRESULTS,
		(void **)&pEnumResults);
	int rtc;

	/* set errors (virus or unscannable item) */
	if (SOPHOS_SUCCEEDED(hr))
	{
		if (hr == SOPHOS_S_OK)
			rtc = 0;
		else
			rtc = ListInfections(pEnumResults, parm, vm, location);
	}
	else
		rtc = savi_scan_error(parm, hr, vm, location);
	
	if (pEnumResults)
		pEnumResults->pVtbl->Release(pEnumResults);
		
	return rtc;
}

static int scan_mime_part(av_parm *parm, MIME *mime, filter_msg *vm)
{
	content_ptr part;
	int rtc = 1; /* cannot mmap */
	
	if (mime_mmap_content(mime, &part) == 0)
	{
		char locbuf[32];
		char const *location = mime_get_filename(mime);
		if (location == NULL)
		{
			location = mime_get_name(mime);
			if (location == NULL)
			{
				sprintf(locbuf, "part #%d", mime_get_position(mime) + 1);
				location = locbuf;
			}
		}
		rtc = scan_memory(parm, vm, location, &part);
		mime_unmap(mime);
	}
	else
		fprintf(stderr, "ERR: avfilter: Cannot mmap MIME part %d: %s (%s)\n",
			mime_get_position(mime),
			mime_get_last_error_string(mime),
			strerror(errno));
	return rtc;
}

static int
scan_using_engine_parser(av_parm *parm, filter_msg *vm)
/*
** scan_malformed, try scanning the mime file directly
*/
{
	content_ptr mmarea;
	int fd = fileno(parm->input_fp), fstat_err;
	struct stat st;
	int rtc = 0;
	
	errno = 0;
	if ((fstat_err = fstat(fd, &st)) == 0 &&
		(mmarea.start = mmap(NULL, mmarea.length = st.st_size,
			PROT_READ, MAP_SHARED, fd, 0)) != MAP_FAILED)
	{
		if (!parm->eng_has_Mime)
			rtc |= set_engine_option(parm->pSAVI, "Mime", parm->verbose, 1);
		if (!parm->eng_has_Base64)
			rtc |= set_engine_option(parm->pSAVI, "Base64", parm->verbose, 1);
		if (scan_memory(parm, vm, "mail file", &mmarea))
			rtc = -1;
		munmap(mmarea.start, mmarea.length);
	}
	else
	{
		fprintf(stderr, "ERR: avfilter: cannot %s mail file: %s\n",
			fstat_err ? "fstat" : "mmap", strerror(errno));
		rtc = -1;
	}
	return rtc;
}

static av_parm* get_parm(fl_parm *fl)
{
	av_parm* parm = (av_parm*) fl_get_parm(fl);
	if (parm == NULL || parm->x55aa5a5a != 0x55aa5a5a)
		abort();
	return parm;
}

static void avfilter(fl_parm *fl)
{
	static char const*const resp_ok = "200 Ok.\n";
	static char const*const resp_bad = "451 system malfunction in avfilter\n";
	/* static char const*const resp_std = "432 Mail filters temporarily unavailable.\n"; */
	
	av_parm *parm = get_parm(fl);
	FILE *fp = parm->input_fp = fl_get_file(fl);
	MIME *mime = mime_new(mime_encoding_not_set, mime_charset_not_set);
	char const *rtc = mime ? read_message(parm, mime, fp) : resp_bad;
	filter_msg vm;
	
	filter_msg_init(&vm);
	
	if (rtc == NULL)
	/*
	** mail message corretly parsed
	*/
	{
		size_t const parts = mime_get_count_all(mime);
		size_t i;
		
		for (i = 0; i < parts; ++i)
		{
			mime_set_position(mime, i);
			if (mime_is_leaf_part(mime) &&
				mime_get_mime_subtype(mime) != mime_text_plain &&
				!mime_is_empty(mime) &&
				scan_mime_part(parm, mime, &vm) != 0)
			{
				rtc = resp_bad;
				break;
			}
		}
	}
	mime_clear(mime);

	if (rtc && *rtc == 0) /* internal mime parser failed */
	{
		rtc = NULL;
		if (scan_using_engine_parser(parm, &vm))
		{
			rtc = resp_bad;
		}
	}
	
	/*
	** examine result
	*/
	if (rtc == NULL)
	{
		if (vm.action >= msg_drop)
		{
			if (fl_drop_message(fl, vm.errmsg) == 0)
			{
				rtc = resp_ok;
			}
		}
					
		if (rtc == NULL && vm.errmsg != NULL)
			rtc = filter_msg_detach(&vm); /* freed on exit */
	}
	filter_msg_clear(&vm);

	if (rtc == NULL)
		rtc = resp_ok;

	fl_pass_message(fl, rtc);
	
#if !defined(NDEBUG)
	if (parm->verbose > 1)
		fprintf(stderr, "avfilter[%lu]: action was %s [%d]\n",
			 (unsigned long) getpid(),
			filter_msg_action_name(vm.action), vm.action);
#endif
}

static int set_savi_scan_error(av_parm *parm, PARSE *p)
/*
** syntax: scan_error num [ '[' action ']' ] = [msg] (was: scan_error num = [msg])
** where action is a string for pass/bounce/drop behaviour
*/
{
	FILE *const report = parm && parm->report ? parm->report : stderr;
	char *err_code = NULL, *msg = NULL;
	int err = 0;
	int tok = get_token(p);
	HRESULT hr = 0; /* compiler happy */
	filter_msg_action action = msg_none;
	int const entry_line = p->line;
	
	if (tok == '_')
	{
		err_code = get_word(p);
		if (err_code)
		{
			tok = get_token_peek(p);
			if (tok == '[')
			{
				int tok2;
				char const *action_w;
				
				get_token(p); // '['
				tok2 = get_token(p);
				action_w = tok2 == '_' ? get_next(p) : NULL;
				if (action_w)
				{
					if (strcmp(action_w, "pass") == 0)
						action = msg_pass;
					else if (strcmp(action_w, "bounce") == 0)
						action = msg_bounce;
					else if (strcmp(action_w, "drop") == 0)
						action = msg_drop;
					else tok2 = 0;
				}
				if (tok2 != '_' || get_token(p) != ']')
				{
					fprintf(report,
						"WARN: avfilter: invalid action code \"%s [%s]\" at line %d in %s\n",
						Psz(err_code), Psz(action_w), p->line, p->fname);
					parm->err = 1;
				}
				tok = get_token_peek(p);
			}
			if (tok == '=')
				tok = get_token(p); // consume '='
			else if (tok == '_')
				err = 2;
			else
				err = 1;

			if (err == 0)
			{
				char *t = NULL;
				
				tok = get_token(p);
				if (tok == '"')
				{
					msg = get_quoted(p);
					tok = get_token(p);
					if (strchr(msg, '\n'))
						err = 3;
				}
				else if (p->line == entry_line)
				{
					msg = get_toeol(p);
					tok = get_token(p);
				}
				else
					msg = "";
				
				hr = strtoul(err_code, &t, 0);
				if (t == NULL || *t != 0)
					// test no repeat also... || SOPHOS_SUCCEEDED(hr))
					err = 1;
			}
		}
	}
	
	if (err || err_code == NULL || msg == NULL)
	{
		fprintf(report,
			"WARN: avfilter: %s for \"scan_error %s%s%s\" at line %d in %s\n",
			err == 2 ? "missing '=' sign" :
				err == 3 ? "multi-line not allowed" :
					"invalid scan_error",
			err == 2 ? msg == NULL || *msg == 0 ? "" : " " : " = ",
			Psz(err_code), Psz(msg), p->line, p->fname);
		if (err == 1)
			tok = get_toeol(p) == NULL ?  -1 : get_token(p);
	}
	else
	{
		int const ndx = parm->savi_se_table_size;
		savi_se *table = (savi_se*)malloc((ndx + 1) * sizeof(savi_se));
		if (table == NULL)
		{
			fprintf(report,
				"ERR: avfilter: cannot allocate %d savi table elements at line %d in %s:\n",
				ndx, p->line, p->fname);
			parm->err = 1;
		}
		if (parm->savi_se_table != NULL)
		{
			memcpy(table, parm->savi_se_table, ndx * sizeof(savi_se));
			free(parm->savi_se_table);
		}
		table[ndx].hr = hr;
		table[ndx].msg = *msg ? msg : NULL;
		table[ndx].action = action > msg_none ? action : *msg ? msg_bounce : msg_pass;
		parm->savi_se_table = table;
		parm->savi_se_table_size = ndx + 1;
	}
	
	return tok;
}

static int set_savi_option(av_parm *parm, PARSE *p)
{
	FILE *const report = parm && parm->report ? parm->report : stderr;
	CISavi3 *const pSAVI = parm->pSAVI;
	char *name = NULL, *value = NULL, *type = NULL;
	int vtype = 3, err = 0;
	int tok = get_token(p);
	
	if (tok == '_')
	{
		name = get_word_delimited(p, "-_", 0, 1);
		if (name)
		{
			tok = get_token(p);
			if (tok == '[')
			{
				type = get_word(p);
				if (type)
				{
					char *t = NULL;
					long l = strtoul(type, &t, 10);
					if (t && *t == 0 && l < 10)
						vtype = (int)l;
					else
						err = 1;
				}
				if ((tok = get_token(p)) == ']')
					tok = get_token(p);
			}
			err |= (tok != '=');
			value = get_toeol(p);
			tok = get_token(p);
		}
	}
	
	if (err || name == NULL || value == NULL)
	{
		fprintf(report, "WARN: avfilter: invalid \"%s\" [%s] = \"%s\" at line %d in %s\n",
			Psz(name), Psz(type), Psz(value), p->line, p->fname);
		parm->err = 1;
	}
	else
	{
		HRESULT hr = pSAVI->pVtbl->SetConfigValue(pSAVI, name, vtype, value);
		if (SOPHOS_FAILED(hr))
		{
			fprintf(report,
				"WARN: avfilter: cannot set %s[%d]=%s at line %d in %s: [0x%08lx]\n",
				name, vtype, value, p->line, p->fname, (unsigned long)hr);
			parm->err = 1;
		}
	}
	
	return tok;
}

int configure_action(av_parm *parm, PARSE *p)
{
	FILE *const report = parm && parm->report ? parm->report : stderr;
	char *name = NULL, *value = NULL;
	filter_msg_action action = msg_none;
	int err = 0;
	int tok = get_token(p);
	
	if (tok == '_')
	{
		name = get_word_delimited(p, "-_/", 0, 1);
		tok = get_token(p);
	}
	else if (tok == '"')
	{
		name = get_quoted(p);
		tok = get_token(p);
	}
	else if (tok == '=')
		name = "default";
	
	if (name && tok == '=')
	{
		value = get_next(p);
		if (value)
		{
			if (stricmp(value, "drop") == 0)
				action = msg_drop;
			else if (stricmp(value, "bounce") == 0)
				action = msg_bounce;
		}
	}
	
	if (name && action > msg_none)
	{
		if (strcmp(name, "default") != 0)
			err = 1;
		
		tok = get_token(p);
	}
	
	if (err || name == NULL || value == NULL || action == msg_none)
	{
		fprintf(report, "WARN: avfilter: invalid action \"%s\" = \"%s\" at line %d in %s\n",
			Psz(name), Psz(value), p->line, p->fname);
		parm->err = 1;
	}
	else
	{
		parm->action_default = action;
	}
	
	return tok;
}


int configure_directory(char **dir, av_parm *parm)
{
	PARSE *const p = parm->parse;
	FILE *const report = parm && parm->report ? parm->report : stderr;
	int tok = get_token(p);
	if (tok == '=')
	{
		char *d = get_toeol(p);
		if (d)
		{
			struct stat s;
			errno = 0;
			if (stat(d, &s) != 0 ||
				!S_ISDIR(s.st_mode) ||
				(S_IRWXU & s.st_mode) != S_IRWXU ||
				(S_IRWXO & s.st_mode) != 0 ||
				s.st_uid != MAILUID)
			{
				static char const err[] = "ERR";
				char const *const w = (errno || !S_ISDIR(s.st_mode)) ? err : "WARN";
				if (w == NULL || parm->verbose > 0)
					fprintf(report,
						"%s: avfilter: unsuitable directory %s: %s in %s line %d\n",
						w, d,
						errno ? strerror(errno) :
							!S_ISDIR(s.st_mode) ? "not a directory" :
							s.st_uid != MAILUID ? "not owned by " MAILUSER :
							(S_IRWXU & s.st_mode) != S_IRWXU ? "not rwx by " MAILUSER:
							(S_IRWXO & s.st_mode) != 0 ? "world accessible" : "uh?",
						p->fname, p->line);
				if (w == err)
				{
					d = NULL;
					parm->err = 1;
				}
			}
			
			if (d)
				*dir = d; /* kept there until parse_clear */
			tok = get_token(p);
		}
		else
		{
			parm->err = 1;
			tok = -1;
		}
	}
	else
	{
		fprintf(report,
			"ERR: avfilter: '=' followed by directory name expected in %s at line %d\n",
			p->fname, p->line);
		parm->err = 1;
	}
	return tok;
}

static int configure(av_parm *parm, FILE *report)
/*
** read the parameter file and set any option
*/
{
	PARSE *const p = parm ? parm->parse : NULL;
	int tok;
	char *next = NULL;

	if (p == NULL)
		return parm && parm->err ? parm->err : -1;

	if (report == NULL)
		report = stderr;

#if !defined(NDEBUG)
	fprintf(report, "avfilter: reading parameters from %s\n", p->fname);
#endif
	
	tok = get_token(p);
	while (tok >= 0)
	{
		if (tok == '_')
		{
			next = get_next(p);
			if (next == NULL)
				break;
		}
		else
		{
			if (tok > 0)
				fprintf(report, "avfilter: Invalid token %c at line %d in %s\n",
					tok, p->line, p->fname);
			break;
		}
	
		if (strcmp(next, "set") == 0)
			tok = set_savi_option(parm, p);
		else if (strcmp(next, "action") == 0)
			tok = configure_action(parm, p);
		else if (strcmp(next, "scan_error") == 0)
			tok = set_savi_scan_error(parm, p);
		else if (strcmp(next, "verbose") == 0)
		{
			parm->verbose = 1;
			if ((tok = get_token(p)) == '=')
			{
				if ((tok = get_token(p)) == '_' &&
					(next = get_next(p)) != NULL)
				{
					char *t = NULL;
					unsigned long l = strtoul(next, &t, 10);
					tok = get_token(p);
					if (t && *t == 0 && l < 16)
						parm->verbose = (int)l;
					else
						tok = -1;
				}
				else
					tok = -1;
			}
		}
		else if (strcmp(next, "scan_malformed") == 0)
		{
			parm->scan_malformed = 1;
			if ((tok = get_token(p)) == '=')
			{
				if ((tok = get_token(p)) == '_' &&
					(next = get_next(p)) != NULL)
				{
					if (next[0] == 'n' || next[0] == 'N' || next[0] == '0')
						parm->scan_malformed = 0;
					tok = get_token(p);
				}
				else
					tok = -1;
			}
		}
		else if (strcmp(next, "save_missed") == 0)
			tok = configure_directory(&parm->save_missed_file, parm);
		else if (strcmp(next, "save_virus") == 0)
			tok = configure_directory(&parm->save_virus_file, parm);
		else if (strcmp(next, "tmp_dir") == 0)
			tok = configure_directory(&parm->tmp_dir, parm);
		else
		{
			fprintf(report, "avfilter: invalid verb \"%s\" at line %d in %s\n",
				next, p->line, p->fname);
			break;
		}
	}

	if (tok != 0)
	{
		fprintf(report, "ERR: avfilter: error after \"%.*s\" parsing %s, line %d\n", 
			MAXWORDLENGTH, Psz(next), Psz(p->fname), p->line);
		parm->err = 1;
	}
	
	if (parm->err == 0)
	{
		if (get_engine_option(parm->pSAVI, "Mime", parm->verbose) == 1)
			parm->eng_has_Mime = 1;
		if (get_engine_option(parm->pSAVI, "Base64", parm->verbose) == 1)
			parm->eng_has_Base64 = 1;

		if (parm->verbose > 1)
			fprintf(report,
				"avfilter: config loaded ok from %s\n", p->fname);
	}
	
	return parm->err;
}

static void open_report_pipe(av_parm *parm)
{
	int const fd = open(AVFILTER_PIPE_FILE, O_WRONLY
#if defined(O_NONBLOCK)
		| O_NONBLOCK
#elif defined(O_NDELAY)
		| O_NDELAY
#else
#error neither O_NONBLOCK nor O_NDELAY defined
#endif
			);
	if (fd < 0)
	{
		/* ENXIO if nobody reading the pipe, assume signal not by avfilter_sig
		** and file was stale */
		char const *const w = errno == ENXIO || errno == ENOENT ? "INFO" : "ERR";
		fprintf(stderr, "%s: avfilter: cannot open %s: %s\n",
			w,
			AVFILTER_PIPE_FILE, strerror(errno));
		parm->report = NULL;
	}
	else if (parm == NULL || (parm->report = fdopen(fd, "w")) == NULL)
	{
		fprintf(stderr, "ERR: avfilter: cannot fdopen %s on fd %d: %s\n",
			AVFILTER_PIPE_FILE, fd, strerror(errno));
		close(fd);
	}
}

static void close_report_pipe(av_parm *parm)
{
	if (parm && parm->report)
	{
		fputs("\nEND_OF_REPORT\n", parm->report);
		fclose(parm->report);
		parm->report = NULL;
	}
}

static void on_sigusr1(fl_parm* fl)
{
	av_parm p;
	av_parm *parm = get_parm(fl);
	int close_pipe = 0, rtc;
	
	if (parm && parm->report == NULL)
	{
		open_report_pipe(parm);
		close_pipe = 1;
	}
	
	if ((rtc = parm_init(&p, parm, parm->config_file)) == 0)
	{
		if (configure(&p, parm->report) == 0)
		{
			p.report = parm->report;
			parm_clear(parm, clear_all_but_SAVI);
			*parm = p;
			fl_set_verbose(fl, p.verbose);
		}
		else
		{
			parm_clear(&p, clear_all_but_SAVI);
			rtc = 1;
		}
	}
	
	if (rtc && parm->report)
	{
		fputs("Reload config failed\n", parm->report);
	}
	
	if (close_pipe)
		close_report_pipe(parm);
}

static void on_sigusr2(fl_parm* fl)
{
	av_parm *parm = get_parm(fl);
	if (parm->report == NULL)
		open_report_pipe(parm);
	SAVI_load_n_report(parm->pSAVI,
		parm->report ? parm->report : stderr,
		parm->verbose, &parm->virus_load);
	close_report_pipe(parm);
}

static void on_sighup(fl_parm *fl)
{
	open_report_pipe(get_parm(fl));
	on_sigusr1(fl);
	on_sigusr2(fl); // closes report pipe
}

static int report_config(av_parm *parm)
{
	int rtc = 0;
	unsigned i;
	int const table_size = parm->savi_se_table_size;
	savi_se const* const table = parm->savi_se_table;
	
	fprintf(stdout,
		"verbose = %d\n"
		"scan_malformed = %d\n"
		"action default = %s\n"
		"# set Mime given = %d\n"
		"# set Base64 given = %d\n"
		"%ssave_virus = %s\n"
		"%ssave_missed = %s\n"
		"%stmp_dir = %s\n"
		"# entries in scan_error: %u\n",
			parm->verbose,
			parm->scan_malformed,
			filter_msg_action_name(parm->action_default),
			parm->eng_has_Mime,
			parm->eng_has_Base64,
			parm->save_virus_file ? "" : "# ",
			parm->save_virus_file ? parm->save_virus_file : "(null)",
			parm->save_missed_file ? "" : "# ",
			parm->save_missed_file ? parm->save_missed_file : "(null)",
			parm->tmp_dir ? "" : "# ",
			parm->tmp_dir ? parm->tmp_dir : "(null)",
			parm->savi_se_table_size);
	
	for (i = 0; i < table_size; ++i)
	{
		savi_se const * const se = &table[i];
		char const *const quotes = se->msg ? "\"" : "";
		fprintf(stdout, "scan_error %#010lx [%s] = %s%s%s\n",
			(unsigned long)(se->hr),
			filter_msg_action_name(se->action),
			quotes, se->msg ? se->msg : "# not set", quotes);
	}
	
	if (parm->verbose)
		rtc = ListConfigs(parm->pSAVI);
	return rtc;
}

#if !defined(NDEBUG)
static void report_config_fl(fl_parm *fl)
{
	report_config(get_parm(fl));
}
#endif


static void write_pid_file(fl_parm *fl)
{
#if !defined(TEST_AVFILTER) // no pid file for tavfilter
	FILE *fp = fopen(AVFILTER_PID_FILE, "w");
	char const *failed_action = NULL;
	if (fp)
	{
		fprintf(fp, "%lu\n", (unsigned long) getpid());
		if ((ferror(fp) | fclose(fp)) != 0)
			failed_action = "write";
		get_parm(fl)->pid_created = 1;
	}
	else
		failed_action = "open";
	if (failed_action)
		fprintf(stderr,
			"ALERT: avfilter: cannot %s %s: %s\n"
			"ALERT: avfilter: %s must contain %lu for signalling virus updates\n",
			failed_action, AVFILTER_PID_FILE, strerror(errno),
			AVFILTER_PID_FILE, (unsigned long) getpid());
#else
	(void)fl;
#endif			
}

static void delete_pid_file(av_parm *parm)
{
#if !defined(TEST_AVFILTER) // no pid file for tavfilter
	if (parm->pid_created &&
		unlink(AVFILTER_PID_FILE) != 0)
			fprintf(stderr, "ERR: avfilter: cannot delete %s: %s\n",
				AVFILTER_PID_FILE, strerror(errno));	
#else
	(void)parm;
#endif			
}

static fl_init_parm functions =
{
	avfilter,
	write_pid_file,
	on_sighup, on_sigusr1, on_sigusr2,
#if !defined(NDEBUG)
	report_config_fl,
#else
	NULL,
#endif
	NULL, NULL, NULL	
};

int main(int argc, char *argv[])
{
	int rtc = 0, i;
	av_parm parm;
	char *config_file = NULL;

	for (i = 1; i < argc; ++i)
	{
		char const *const arg = argv[i];
		
		if (strcmp(arg, "-f") == 0)
		{
			config_file = ++i < argc ? argv[i] : NULL;
		}
		else if (strcmp(arg, "--version") == 0)
		{
			puts(PACKAGE_STRING);
			return 0;
		}
		else if (strcmp(arg, "--config") == 0)
			rtc = 1;
		else if (strcmp(arg, "--help") == 0)
		{
			fputs("avfilter command line args:\n"
			/*  12345678901234567890123456 */
				"  --help                  print this stuff and exit\n"
				"  --version               print version string and exit\n"
				"  -f config-file          specify alternate config file\n"
				"  --config                display values from config file\n",
					stdout);
			fl_main(NULL, NULL, argc - i + 1, argv + i - 1, 0);
			return 0;
		}
	}
	
	if (parm_init(&parm, NULL, config_file) ||
		configure(&parm, stderr))
			return 1;

	if (SAVI_load_n_report(parm.pSAVI, stderr,
		parm.verbose, &parm.virus_load))
			rtc = 1;
	else if (rtc)
		rtc = report_config(&parm);
	else
		rtc = fl_main(&functions, &parm, argc, argv, parm.verbose);

	delete_pid_file(&parm);
	parm_clear(&parm, clear_all);
	return rtc;
}
