/*
** filter_msg.h - written in milano by vesely on 5 dec 2002
**
** This is included in avfilter.c
** Extension was "c" for ease of testing,
** now there is a separate tfilter_msg.c file
*/
/*
** Copyright (c) 1999, 2000, 2001, 2002, 2003, 2004 Alessandro Vesely
** All rights reserved. see COPYING
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#if !defined(FILTER_MSG_H_INCLUDED)

typedef enum filter_msg_action
{
	msg_none, msg_pass, msg_bounce, msg_drop
} filter_msg_action;

typedef struct filter_msg_struct
{
	char *errmsg,   // the multi-line error message
		*lasterrmsg; // an internal pointer used for adjustments
	size_t errmsglen;
	filter_msg_action action;
	unsigned short code;
	char code_str[4];
} filter_msg;

static char const*filter_msg_action_name(filter_msg_action action)
{
	static char const *const action_name[] =
		/* 0       1       2         3       4    */
		{ "none", "pass", "bounce", "drop", "INVALID ACTION"};
	return action_name[(unsigned)action < 4U ? action : 4U];
}

static void set_code_str(filter_msg *v)
{
	int code = v->code;
	char buf[8];
	if (code < 100 || code > 999)
		code = v->code = 554;
	sprintf(buf, "%d", code);
	strncpy(v->code_str, buf, sizeof v->code_str);
	v->code_str[sizeof v->code_str - 1] = 0;
}

static void filter_msg_init(filter_msg *v)
{
	memset(v, 0, sizeof *v);
	set_code_str(v);
}

static void filter_msg_clear(filter_msg *v)
{
	if (v->errmsg)
	{
		free(v->errmsg);
		v->errmsg = v->lasterrmsg = NULL;
	}
}

static char *filter_msg_detach(filter_msg *v)
/*
** get the message and the responsibility to free it
*/
{
	char *errmsg = v->errmsg;
	v->errmsg = v->lasterrmsg = NULL;
	return errmsg;
}

static int filter_msg_add(filter_msg *v,
	char const *name, char const *location, filter_msg_action action)
{
	static char const fmt[] = "%3.3s %s \"%s\" in %s\n";
	int const virus = action >= msg_drop ? 1 : 0;
	size_t required = strlen(name) + strlen(location) + (sizeof fmt + 5);
	if (v->lasterrmsg)
		v->lasterrmsg[3] = '-';
	if ((v->lasterrmsg = (char*)malloc(required + v->errmsglen)) == NULL)
		return 1;
	if (v->errmsg)
	{
		memcpy(v->lasterrmsg, v->errmsg, v->errmsglen);
		free(v->errmsg);
	}
	v->errmsg = v->lasterrmsg;
	v->lasterrmsg += v->errmsglen;
	v->errmsglen += sprintf(v->lasterrmsg, fmt,
		v->code_str, virus ? "Virus" : "Error", name, location);
	
	if (action > v->action)
		v->action = action;
	return 0;
}

#define FILTER_MSG_H_INCLUDED 1
#endif
